#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

// globals for rendering
int fullscreen_mode=0;
int width=800, height=600;
int saved_width, saved_height;
int oldmousex=0, oldmousey=0;
enum {DONOTHING, ROTATE, ZOOM, MOVETARGET} mousemode=DONOTHING;
float camdist=6;
const float NEARCLIP_FACTOR=0.1, FARCLIP_FACTOR=10;
float nearclip, farclip;
float target[3]={0, 1, 0};
float view_target_timer=0;
float old_time=0;
const float fovy=15;
float heading=M_PI/3, pitch=0;
float campos[3];
int file_count=0;
int skip=1;

// globals for data
int frame=0;
const char *directory=0;
int nx=0, ny=0, nz=0;
float *s=0, *block=0;
#define S(i,j,k) (s[((i)*ny+(j))*nz+(k)])
#define BLOCK(i,j,k) (block[((i)*ny+(j))*nz+(k)])

// prototypes
void read_frame(int new_frame);
void find_reasonable_camera();
void handle_display();
void handle_reshape(int w, int h);
void handle_keypress(unsigned char key, int x, int y);
void handle_click(int button, int state, int x, int y);
void handle_drag(int x, int y);
void handle_idle();
void toggle_fullscreen();
void compute_camera_position();
void display_smoke();
void display_target();
void save_screen(const char *filename);

// definitions
int main(int argc, char *argv[])
{
   // initialize GLUT stuff 
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA|GLUT_DEPTH);
   glutInitWindowSize(width,height);
   saved_width=width;
   saved_height=height;
   glutCreateWindow("smoke viewer (hw6)");
   glutDisplayFunc(handle_display);
   glutReshapeFunc(handle_reshape);
   glutKeyboardFunc(handle_keypress);
   glutMouseFunc(handle_click);
   glutMotionFunc(handle_drag);
   glutIdleFunc(handle_idle);

   // initialize OpenGL stuff
   glEnable(GL_DEPTH_TEST);
   glClearColor(0.5, 0.5, 1, 0);
   glClearDepth(1);
   glPixelStorei(GL_PACK_ALIGNMENT, 1);
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   glEnable(GL_BLEND);
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

   if(argc<2){
      fprintf(stderr, "Expecting <directory> as argument\n");
      return 1;
   }
   directory=argv[1];

   // initialize first frame
   read_frame(0);
   find_reasonable_camera();
   nearclip=NEARCLIP_FACTOR*camdist;
   farclip=FARCLIP_FACTOR*camdist;
   compute_camera_position();

   // let's go
   old_time=clock()/(float)CLOCKS_PER_SEC;
   glutPostRedisplay();
   glutMainLoop();
   return 0;
}

void read_frame(int new_frame)
{
   char *filename;
   FILE *fp;
   int i, j, k;

   if(new_frame<0) return;
   filename=malloc(strlen(directory)+30);
   sprintf(filename, "%s/smoke%04d", directory, new_frame);
   fp=fopen(filename, "rt");
   if(fp){
      int newnx, newny, newnz;
      fscanf(fp, "%d %d %d", &newnx, &newny, &newnz);
      if(newnx!=nx || newny!=ny || newnz!=nz || s==0){
         free(s); free(block);
         nx=newnx;
         ny=newny;
         nz=newnz;
         s=(float*)malloc(nx*ny*nz*sizeof(float));
         block=(float*)malloc(nx*ny*nz*sizeof(float));
      }
      frame=new_frame;
      for(i=0; i<nx*ny*nz; ++i){
         fscanf(fp, "%f", s+i);
      }
      fclose(fp);

      // now compute light coming through smoke from above at an angle
      for(i=0; i<nx; ++i){
         for(j=ny-1; j>=0; --j){
            for(k=0; k<nz; ++k){
               BLOCK(i,j,k)=0;
               if(j<ny-1)               BLOCK(i,j,k)+=0.05*(BLOCK(i,j+1,k)+S(i,j+1,k));
               if(i>0 && j<ny-1)        BLOCK(i,j,k)+=0.025*(BLOCK(i-1,j+1,k)+S(i-1,j+1,k));
               if(j<ny-1 && k>0)        BLOCK(i,j,k)+=0.025*(BLOCK(i,j+1,k-1)+S(i,j+1,k-1));
               if(i>0 && j<ny-1 && k>0) BLOCK(i,j,k)+=0.9*(BLOCK(i-1,j+1,k-1)+S(i-1,j+1,k-1));
            }
         }
      }
      for(i=0; i<nx; ++i)
         for(j=0; j<ny; ++j)
            for(k=0; k<nz; ++k)
               BLOCK(i,j,k)=(0.8*exp(-500*BLOCK(i,j,k)/(nx+nz))+0.2);
   }
   free(filename);
}

void find_reasonable_camera()
{
   target[0]=nx/2;
   target[1]=ny/2;
   target[2]=nz/2;
   camdist=2*(nx+ny+nz);
}

void handle_display()
{
   glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

   glMatrixMode (GL_MODELVIEW);
   glLoadIdentity();
   gluLookAt(campos[0], campos[1], campos[2], target[0], target[1], target[2], 0, 1, 0);

   display_smoke();
   if(view_target_timer>0) display_target();

   glutSwapBuffers();
}

void handle_reshape(int w, int h)
{
   width=w;
   height=h;
   glViewport(0, 0, (GLsizei)width, (GLsizei)height);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(fovy, width/(float)height, nearclip, farclip);
   glutPostRedisplay();
}

void handle_keypress(unsigned char key, int x, int y)
{
   switch(key){
   case '.':
      read_frame(frame+1);
      break;
   case ',':
      read_frame(frame-1);
      break;
   case '>':
      read_frame(frame+10);
      break;
   case '<':
      read_frame(frame-10);
      break;
   case 'f':
      toggle_fullscreen();
      break;
   case 'd':
      if(skip==1){
         skip=(nx*nz)/(25*25);
         if(skip<=0) skip=1;
      } else skip=1;
      break;
   case 'q':
      exit(0);
   case 's':
      {
         char filename[30];
         sprintf(filename, "hw6_%04d.ppm", file_count);
         save_screen(filename);
         ++file_count;
      }
      break;
   default:
      return;
   }
   glutPostRedisplay();
}

void handle_click(int button, int state, int x, int y)
{
   int mods=glutGetModifiers();
   if(state==GLUT_UP) mousemode = DONOTHING;
   else if(mods&GLUT_ACTIVE_SHIFT){
      view_target_timer=1;
      mousemode=MOVETARGET;
   }else{
      if(button==GLUT_RIGHT_BUTTON || (button==GLUT_LEFT_BUTTON && mods&GLUT_ACTIVE_CTRL))
         mousemode=ZOOM;
      else if(button==GLUT_LEFT_BUTTON)
         mousemode=ROTATE;
   }
   oldmousex=x;
   oldmousey=y;
}

void handle_drag(int x, int y)
{
   switch(mousemode){
   case ROTATE:
      heading+=0.007*(oldmousex-x);
      if(heading<-M_PI) heading+=2*M_PI;
      else if(heading>M_PI) heading-=2*M_PI;
      pitch+=0.007*(oldmousey-y);
      if(pitch<-.48*M_PI) pitch=-.48*M_PI;
      else if(pitch>.48*M_PI) pitch=.48*M_PI;
      break;
   case ZOOM:
      camdist*=pow(1.01, y-oldmousey);
      nearclip=NEARCLIP_FACTOR*camdist;
      farclip=FARCLIP_FACTOR*camdist;
      break;
   case MOVETARGET:
      target[0]-=(0.002*camdist)*cos(heading)*(oldmousex-x);
      target[1]-=(0.002*camdist)*(oldmousey-y);
      target[2]+=(0.002*camdist)*sin(heading)*(oldmousex-x);
      break;
   case DONOTHING:
      break;
   }
   oldmousex=x;
   oldmousey=y;
   compute_camera_position();
   glutPostRedisplay();
}

void handle_idle()
{
   float new_time=clock()/(float)CLOCKS_PER_SEC;
   float delta_time=new_time-old_time;
   int target_turned_off=0;
   if(delta_time<0) delta_time=0; // this can happen due to roll-over
   old_time=new_time;
   if(view_target_timer>0){
      view_target_timer-=delta_time;
      if(view_target_timer<=0){
         target_turned_off=1;
         view_target_timer=0;
      }
   }
   if(target_turned_off)
      glutPostRedisplay();
}

void toggle_fullscreen()
{
   if(fullscreen_mode){
      fullscreen_mode=0;
      width=saved_width;
      height=saved_height;
      glutReshapeWindow(saved_width, saved_height);
   }else{
      fullscreen_mode=1;
      saved_width=width;
      saved_height=height;
      glutFullScreen();
   }
}

void compute_camera_position()
{
   float ch=cos(heading), sh=sin(heading), cp=cos(pitch), sp=sin(pitch);
   campos[0]=target[0]-camdist*sh*cp;
   campos[1]=target[1]-camdist*sp;
   campos[2]=target[2]-camdist*ch*cp;
}

static inline void smoke_vertex(i,j,k)
{
   float a=15*skip*S(i,j,k)/(nx+nz), g=BLOCK(i,j,k);
   glColor4f(g, g, g, a);
   glVertex3i(i,j,k);
}

void draw_x_reverse()
{
   int i, j, k;
   for(i=nx-1; i>=0; i-=skip){
      for(j=0; j<ny-skip; j+=skip){
         glBegin(GL_QUAD_STRIP);
         for(k=0; k<nz; k+=skip){
            smoke_vertex(i,j+skip,k);
            smoke_vertex(i,j,k);
         }
         glEnd();
      }
   }
}

void draw_x_forward()
{
   int i, j, k;
   for(i=0; i<nx; i+=skip){
      for(j=0; j<ny-skip; j+=skip){
         glBegin(GL_QUAD_STRIP);
         for(k=0; k<nz; k+=skip){
            smoke_vertex(i,j,k);
            smoke_vertex(i,j+skip,k);
         }
         glEnd();
      }
   }
}

void draw_y_reverse()
{
   int i, j, k;
   for(j=ny-1; j>=0; j-=skip){
      for(i=0; i<nx-skip; i+=skip){
         glBegin(GL_QUAD_STRIP);
         for(k=0; k<nz; k+=skip){
            smoke_vertex(i,j,k);
            smoke_vertex(i+skip,j,k);
         }
         glEnd();
      }
   }
}

void draw_y_forward()
{
   int i, j, k;
   for(j=0; j<ny; j+=skip){
      for(i=0; i<nx-skip; i+=skip){
         glBegin(GL_QUAD_STRIP);
         for(k=0; k<nz; k+=skip){
            smoke_vertex(i,j,k);
            smoke_vertex(i+skip,j,k);
         }
         glEnd();
      }
   }
}

void draw_z_reverse()
{
   int i, j, k;
   for(k=nz-1; k>=0; k-=skip){
      for(i=0; i<nx-skip; i+=skip){
         glBegin(GL_QUAD_STRIP);
         for(j=0; j<ny; j+=skip){
            smoke_vertex(i,j,k);
            smoke_vertex(i+skip,j,k);
         }
         glEnd();
      }
   }
}

void draw_z_forward()
{
   int i, j, k;
   for(k=0; k<nz; k+=skip){
      for(i=0; i<nx-skip; i+=skip){
         glBegin(GL_QUAD_STRIP);
         for(j=0; j<ny; j+=skip){
            smoke_vertex(i,j,k);
            smoke_vertex(i+skip,j,k);
         }
         glEnd();
      }
   }
}

void display_smoke()
{
   float dx=target[0]-campos[0], dy=target[1]-campos[1], dz=target[2]-campos[2];
   if(fabs(dx)>=fabs(dy) && fabs(dx)>=fabs(dz)){
      if(dx>0) draw_x_reverse();
      else draw_x_forward();
   }else if(fabs(dz)>=fabs(dy)){
      if(dz>0) draw_z_reverse();
      else draw_z_forward();
   }else{
      if(dy>0) draw_y_reverse();
      else draw_y_forward();
   }
}

void display_target()
{
   float logcd=log(camdist)/log(20.);
   float smallsize=pow(20.0, floor(logcd-.5));
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glTranslatef(target[0], target[1], target[2]);
   glColor4f(1, 1, 1, 1);
   glutWireCube(smallsize);
   glutWireCube(20*smallsize);
   glPopMatrix();
}

void save_screen(const char *filename)
{
   FILE *fp;
   GLubyte *image_buffer;

   // get the pixels
   image_buffer=malloc(3*width*height*sizeof(GLubyte));
   glReadBuffer(GL_FRONT);
   glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, image_buffer);
  
   // write them to the file
   fp=fopen(filename, "wb");
   if(fp){
      int i;
      fprintf(fp, "P6\n%d %d 255\n", width, height);
      for(i=1; i<=height; ++i){
         fwrite(image_buffer+3*width*(height-i), 1, 3*width, fp);
      }
   }
   
   // clean up
   free(image_buffer);
}

